//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/13
// 
//   Description:  Implements many of the operations for 8-bit variables.
//
//   Notes:
//
//      Some handy equations for synthesizing the missing intrinsics.
//
//      a <= b => (a - min(a,b)) == 0
//      a >= b => (b - min(a,b)) == 0
//      a >  b => (a - min(a,b)) != 0
//      a <  b => (b - min(a,b)) != 0
//
//      Since we don't have a NEQ intrinsice we XOR with -1.
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "SSELibrary.h"
#include <immintrin.h>

#pragma region CmpEQ
SSELIBRARY_API void SSE_CmpEQ_8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_cmpeq_epi8(m_dst, m_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void AVX2_CmpEQ_8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_cmpeq_epi8(m_dst, m_src); // AVX2 intrinsic
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_CmpEQ_8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (unsigned char)(*pdst == *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}
#pragma endregion // CmpEQ

#pragma region CmpGT
SSELIBRARY_API void SSE_CmpGT_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_cmpgt_epi8(m_dst, m_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void AVX2_CmpGT_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_cmpgt_epi8(m_dst, m_src); // AVX2 intrinsic
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_CmpGT_I8(char* dst, char* src, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (char)(*pdst > *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}

SSELIBRARY_API void SSE_CmpGT_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    __m128i ZERO = _mm_setzero_si128();
    __m128i FF = _mm_set1_epi8(-1);

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);

        __m128i result = _mm_min_epu8(m_dst, m_src);
        result = _mm_subs_epu8(m_dst, result);
        result = _mm_cmpeq_epi8(result,ZERO);
        result = _mm_xor_si128(result, FF);

        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void AVX2_CmpGT_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i ZERO = _mm256_setzero_si256();
    __m256i m_FF = _mm256_set1_epi8(-1);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);

        __m256i result = _mm256_min_epu8(m_dst, m_src);
        result = _mm256_subs_epu8(m_dst, result);
        result = _mm256_cmpeq_epi8(result,ZERO);
        result = _mm256_xor_si256(result, m_FF);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_CmpGT_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (unsigned char)(*pdst > *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}
#pragma endregion // CmpGT

#pragma region CmpLT
SSELIBRARY_API void SSE_CmpLT_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_cmplt_epi8(m_dst, m_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void AVX2_CmpLT_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i ZERO = _mm256_setzero_si256();
    __m256i m_FF = _mm256_set1_epi8(-1);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);

        __m256i result = _mm256_min_epi8(m_dst, m_src);
        result = _mm256_subs_epi8(m_dst, result);
        result = _mm256_cmpeq_epi8(result,ZERO);
        result = _mm256_xor_si256(result, m_FF);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_CmpLT_I8(char* dst, char* src, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (char)(*pdst < *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}

SSELIBRARY_API void SSE_CmpLT_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    __m128i ZERO = _mm_setzero_si128();
    __m128i FF = _mm_set1_epi8(-1);

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);

        __m128i result = _mm_min_epu8(m_dst, m_src);
        result = _mm_subs_epu8(m_src, result);
        result = _mm_cmpeq_epi8(result,ZERO);
        result = _mm_xor_si128(result, FF);

        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void AVX2_CmpLT_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i ZERO = _mm256_setzero_si256();
    __m256i m_FF = _mm256_set1_epi8(-1);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);

        __m256i result = _mm256_min_epu8(m_dst, m_src);
        result = _mm256_subs_epu8(m_dst, result);
        result = _mm256_cmpeq_epi8(result,ZERO);
        result = _mm256_xor_si256(result, m_FF);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_CmpLT_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (unsigned char)(*pdst < *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}
#pragma endregion CmpLT

#pragma region CmpLTE
SSELIBRARY_API void SSE_CmpLTE_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    __m128i ZERO = _mm_setzero_si128();

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);

        __m128i result = _mm_min_epu8(m_dst, m_src);
        result = _mm_subs_epu8(m_dst, result);
        result = _mm_cmpeq_epi8(result,ZERO);

        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_CmpLTE_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (unsigned char)(*pdst <= *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}
SSELIBRARY_API void AVX2_CmpLTE_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i ZERO = _mm256_setzero_si256();
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);

        __m256i result = _mm256_min_epu8(m_dst, m_src);
        result = _mm256_subs_epu8(m_dst, result);
        result = _mm256_cmpeq_epi8(result,ZERO);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void SSE_CmpLTE_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;
    __m128i ZERO = _mm_setzero_si128();

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);

        __m128i result = _mm_min_epi8(m_dst, m_src);
        result = _mm_subs_epi8(m_dst, result);
        result = _mm_cmpeq_epi8(result,ZERO);

        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_CmpLTE_I8(char* dst, char* src, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (char)(*pdst <= *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}
SSELIBRARY_API void AVX2_CmpLTE_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i ZERO = _mm256_setzero_si256();
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);

        __m256i result = _mm256_min_epu8(m_dst, m_src);
        result = _mm256_subs_epi8(m_dst, result);
        result = _mm256_cmpeq_epi8(result,ZERO);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
#pragma endregion // CmpLTE

#pragma region CmpGTE
SSELIBRARY_API void SSE_CmpGTE_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;
    __m128i ZERO = _mm_setzero_si128();

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);

        __m128i result = _mm_min_epu8(m_dst, m_src);
        result = _mm_subs_epu8(m_src, result);
        result = _mm_cmpeq_epi8(result,ZERO);

        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_CmpGTE_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (unsigned char)(*pdst >= *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}
SSELIBRARY_API void AVX2_CmpGTE_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i ZERO = _mm256_setzero_si256();
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);

        __m256i result = _mm256_min_epu8(m_dst, m_src);
        result = _mm256_subs_epu8(m_src, result);
        result = _mm256_cmpeq_epi8(result,ZERO);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void SSE_CmpGTE_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    __m128i ZERO = _mm_setzero_si128();

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);

        __m128i result = _mm_min_epi8(m_dst, m_src);
        result = _mm_subs_epi8(m_src, result);
        result = _mm_cmpeq_epi8(result,ZERO);

        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_CmpGTE_I8(char* dst, char* src, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
    for (; bytes > 0; bytes--)
	{
        *pdst = (char)(*pdst >= *psrc ? 0xFF : 0x0);
        psrc++;
        pdst++;
	}
}
SSELIBRARY_API void AVX2_CmpGTE_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i ZERO = _mm256_setzero_si256();
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);

        __m256i result = _mm256_min_epu8(m_dst, m_src);
        result = _mm256_subs_epi8(m_src, result);
        result = _mm256_cmpeq_epi8(result,ZERO);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
#pragma endregion // CmpGTE

