//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/13
// 
//   Description:  Implement the bitwise operations - AND, OR, XOR, ANDNOT
//
//   Notes:
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "SSELibrary.h"
#include <immintrin.h>

//
// Simple call to make up front to detect any missing DLL issues.
//
SSELIBRARY_API bool IsPresent()
{
    return true;
}

static bool _init = false;
static bool _sse3 = false;
static bool _ssse3 = false;
static bool _sse41 = false;
static bool _sse42 = false;
static bool _avx = false;
static bool _avx2 = false;
static bool _avx512 = false;

void Init( bool force = false)
{
    if ( force == true || _init == false )
    {
        _sse3 = Supports_SSE3();
        _ssse3 = Supports_SSSE3();
        _sse41 = Supports_SSE41();
        _sse42 = Supports_SSE42();
        _avx = Supports_AVX1();
		_avx2 = Supports_AVX2();
        _avx512 = Supports_AVX512();
        _init = true;
    }
}

SSELIBRARY_API void DisableDispatchFeatures(
          bool avx512 = false
        , bool avx2 = false
        , bool avx = false
        , bool sse42 = false
        , bool sse41 = false
        , bool ssse3 = false
        , bool sse3 = false
    )
{
    Init();
    if (avx512) _avx512 = false;
    if (avx2) _avx2 = false;
    if (avx) _avx = false;
    if (sse42) _sse42 = false;
    if (sse41) _sse41 = false;
    if (ssse3) _ssse3 = false;
    if (sse3) _sse3 = false;
}

SSELIBRARY_API void EnableDispatchFeatures(
          bool avx512 = false
        , bool avx2 = false
        , bool avx = false
        , bool sse42 = false
        , bool sse41 = false
        , bool ssse3 = false
        , bool sse3 = false
    )
{
    Init();
    if (avx512) _avx512 = true;
    if (avx2) _avx2 = true;
    if (avx) _avx = true;
    if (sse42) _sse42 = true;
    if (sse41) _sse41 = true;
    if (ssse3) _ssse3 = true;
    if (sse3) _sse3 = true;
}


typedef void (*FunctionType1) (char* src, size_t bytes);
typedef void (*FunctionType2) (char* dst, size_t bytes, int val);
typedef void (*FunctionType3) (char* dst, char* src, size_t bytes);
typedef void (*FunctionType4) (char* dst, char* src, char* mask, size_t bytes);

#pragma region MEMSET
FunctionType2 FastestMEMSET8 = NULL;
SSELIBRARY_API void MEMSET8(char* dst, size_t bytes, int val)
{
    if(FastestMEMSET8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestMEMSET8 = SSE_MEMSET8; }
        else if (_avx ) { FastestMEMSET8 = AVX_MEMSET8; }
        else { FastestMEMSET8 = CPP_MEMSET8; } // custom CPU implementation
    }
    FastestMEMSET8(dst,bytes,val);
}
#pragma endregion

#pragma region LogicalOperators
FunctionType3 FastestAND = NULL;
SSELIBRARY_API void AND(char* dst, char* src, size_t bytes)
{
    if(FastestAND == NULL)
    {
        // They're all roughly equal with AVX_AND being the slowest.
        Init();
        if ( _sse3 ) { FastestAND = SSE_AND; }
        else if (_avx2 ) { FastestAND = AVX2_AND; }
        else if (_avx ) { FastestAND = AVX_AND; }
        else { FastestAND = CPP_AND; } // custom CPU implementation
    }
    FastestAND(dst,src,bytes);
}

FunctionType3 FastestOR = NULL;
SSELIBRARY_API void OR(char* dst, char* src, size_t bytes)
{
    if(FastestOR == NULL)
    {
        Init();
        if ( _sse3 ) { FastestOR = SSE_OR; }
        else if (_avx2 ) { FastestOR = AVX2_OR; }
        else if (_avx ) { FastestOR = AVX_OR; }
        else { FastestOR = CPP_OR; } // custom CPU implementation
    }
    FastestOR(dst,src,bytes);
}

FunctionType3 FastestXOR = NULL;
SSELIBRARY_API void XOR(char* dst, char* src, size_t bytes)
{
    if(FastestXOR == NULL)
    {
        Init();
        if ( _sse3 ) { FastestXOR = SSE_XOR; }
        else if (_avx2 ) { FastestXOR = AVX2_XOR; }
        else if (_avx ) { FastestXOR = AVX_XOR; }
        else { FastestXOR = CPP_XOR; } // custom CPU implementation
    }
    FastestXOR(dst,src,bytes);
}

FunctionType3 FastestANDNOT = NULL;
SSELIBRARY_API void ANDNOT(char* dst, char* src, size_t bytes)
{
    if(FastestANDNOT == NULL)
    {
        Init();
        if ( _sse3 ) { FastestANDNOT = SSE_ANDNOT; }
        else if (_avx2 ) { FastestANDNOT = AVX2_ANDNOT; }
        else if (_avx ) { FastestANDNOT = AVX_ANDNOT; }
        else { FastestANDNOT = CPP_ANDNOT; } // custom CPU implementation
    }
    FastestANDNOT(dst,src,bytes);
}
#pragma endregion

#pragma region Adds_Subs
FunctionType3 FastestAdds_U8 = NULL;
SSELIBRARY_API void Adds_U8(char* dst, char* src, size_t bytes)
{
    if(FastestAdds_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestAdds_U8 = SSE_Adds_U8; }
        else if (_avx2 ) { FastestAdds_U8 = AVX2_Adds_U8; }
        else { FastestAdds_U8 = CPP_Adds_U8; } // custom CPU implementation
    }
    FastestAdds_U8(dst,src,bytes);
}

FunctionType3 FastestAdds_I8 = NULL;
SSELIBRARY_API void Adds_I8(char* dst, char* src, size_t bytes)
{
    if(FastestAdds_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestAdds_I8 = SSE_Adds_I8; }
        else if (_avx2 ) { FastestAdds_I8 = AVX2_Adds_I8; }
        else { FastestAdds_I8 = CPP_Adds_I8; } // custom CPU implementation
    }
    FastestAdds_I8(dst,src,bytes);
}

FunctionType3 FastestSubs_U8 = NULL;
SSELIBRARY_API void Subs_U8(char* dst, char* src, size_t bytes)
{
    if(FastestSubs_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestSubs_U8 = SSE_Subs_U8; }
        else if (_avx2 ) { FastestSubs_U8 = AVX2_Subs_U8; }
        else { FastestSubs_U8 = CPP_Subs_U8; } // custom CPU implementation
    }
    FastestSubs_U8(dst,src,bytes);
}

FunctionType3 FastestSubs_I8 = NULL;
SSELIBRARY_API void Subs_I8(char* dst, char* src, size_t bytes)
{
    if(FastestSubs_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestSubs_I8 = SSE_Subs_I8; }
        else if (_avx2 ) { FastestSubs_I8 = AVX2_Subs_I8; }
        else { FastestSubs_I8 = CPP_Subs_I8; } // custom CPU implementation
    }
    FastestSubs_I8(dst,src,bytes);
}
#pragma endregion

#pragma region Add_Sub
FunctionType3 FastestSub_I8 = NULL;
SSELIBRARY_API void Sub_I8(char* dst, char* src, size_t bytes)
{
    if(FastestSub_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestSub_I8 = SSE_Sub_I8; }
        else if (_avx2 ) { FastestSub_I8 = AVX2_Sub_I8; }
        else { FastestSub_I8 = CPP_Sub_I8; } // custom CPU implementation
    }
    FastestSub_I8(dst,src,bytes);
}

SSELIBRARY_API void Sub_U8(char* dst, char* src, size_t bytes)
{
	Sub_I8(dst, src, bytes);
}
FunctionType3 FastestAdd_I8 = NULL;
SSELIBRARY_API void Add_I8(char* dst, char* src, size_t bytes)
{
    if(FastestAdd_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestAdd_I8 = SSE_Add_I8; }
        else if (_avx2 ) { FastestAdd_I8 = AVX2_Add_I8; }
        else { FastestAdd_I8 = CPP_Add_I8; } // custom CPU implementation
    }
    FastestAdd_I8(dst,src,bytes);
}
SSELIBRARY_API void Add_U8(char* dst, char* src, size_t bytes)
{
	Add_I8(dst, src, bytes);
}
#pragma endregion

#pragma region AVG
FunctionType3 FastestAvg_U8 = NULL;
SSELIBRARY_API void Avg_U8(char* dst, char* src, size_t bytes)
{
    if(FastestAvg_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestAvg_U8 = SSE_Avg_U8; }
        else if (_avx2 ) { FastestAvg_U8 = AVX2_Avg_U8; }
        else { FastestAvg_U8 = CPP_Avg_U8; } // custom CPU implementation
    }
    FastestAvg_U8(dst,src,bytes);
}
#pragma endregion

#pragma region CmpEQ
FunctionType3 FastestCmpEQ_8 = NULL;
SSELIBRARY_API void CmpEQ_8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpEQ_8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpEQ_8 = SSE_CmpEQ_8; }
        else if (_avx2 ) { FastestCmpEQ_8 = AVX2_CmpEQ_8; }
        else { FastestCmpEQ_8 = CPP_CmpEQ_8; } // custom CPU implementation
    }
    FastestCmpEQ_8(dst,src,bytes);
}
#pragma endregion

#pragma region CmpGT
FunctionType3 FastestCmpGT_I8 = NULL;
SSELIBRARY_API void CmpGT_I8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpGT_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpGT_I8 = SSE_CmpGT_I8; }
        else if (_avx2 ) { FastestCmpGT_I8 = AVX2_CmpGT_I8; }
        else { FastestCmpGT_I8 = CPP_CmpGT_I8; } // custom CPU implementation
    }
    FastestCmpGT_I8(dst,src,bytes);
}

FunctionType3 FastestCmpGT_U8 = NULL;
SSELIBRARY_API void CmpGT_U8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpGT_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpGT_U8 = SSE_CmpGT_U8; }
        else if (_avx2 ) { FastestCmpGT_U8 = AVX2_CmpGT_U8; }
        else { FastestCmpGT_U8 = CPP_CmpGT_U8; } // custom CPU implementation
    }
    FastestCmpGT_U8(dst,src,bytes);
}
#pragma endregion

#pragma region CmpLT
FunctionType3 FastestCmpLT_I8 = NULL;
SSELIBRARY_API void CmpLT_I8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpLT_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpLT_I8 = SSE_CmpLT_I8; }
        else if (_avx2 ) { FastestCmpLT_I8 = AVX2_CmpLT_I8; }
        else { FastestCmpLT_I8 = CPP_CmpLT_I8; } // custom CPU implementation
    }
    FastestCmpLT_I8(dst,src,bytes);
}

FunctionType3 FastestCmpLT_U8 = NULL;
SSELIBRARY_API void CmpLT_U8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpLT_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpLT_U8 = SSE_CmpLT_U8; }
        else if (_avx2 ) { FastestCmpLT_U8 = AVX2_CmpLT_U8; }
        else { FastestCmpLT_U8 = CPP_CmpLT_U8; } // custom CPU implementation
    }
    FastestCmpLT_U8(dst,src,bytes);
}
#pragma endregion

#pragma region CmpLTE
FunctionType3 FastestCmpLTE_U8 = NULL;
SSELIBRARY_API void CmpLTE_U8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpLTE_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpLTE_U8 = SSE_CmpLTE_U8; }
        else if (_avx2 ) { FastestCmpLTE_U8 = AVX2_CmpLTE_U8; }
        else { FastestCmpLTE_U8 = CPP_CmpLTE_U8; } // custom CPU implementation
    }
    FastestCmpLTE_U8(dst,src,bytes);
}

FunctionType3 FastestCmpLTE_I8 = NULL;
SSELIBRARY_API void CmpLTE_I8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpLTE_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpLTE_I8 = SSE_CmpLTE_I8; }
        else if (_avx2 ) { FastestCmpLTE_I8 = AVX2_CmpLTE_I8; }
        else { FastestCmpLTE_I8 = CPP_CmpLTE_I8; } // custom CPU implementation
    }
    FastestCmpLTE_I8(dst,src,bytes);
}
#pragma endregion

#pragma region CmpGTE
FunctionType3 FastestCmpGTE_U8 = NULL;
SSELIBRARY_API void CmpGTE_U8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpGTE_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpGTE_U8 = SSE_CmpGTE_U8; }
        else if (_avx2 ) { FastestCmpGTE_U8 = AVX2_CmpGTE_U8; }
        else { FastestCmpGTE_U8 = CPP_CmpGTE_U8; } // custom CPU implementation
    }
    FastestCmpGTE_U8(dst,src,bytes);
}

FunctionType3 FastestCmpGTE_I8 = NULL;
SSELIBRARY_API void CmpGTE_I8(char* dst, char* src, size_t bytes)
{
    if(FastestCmpGTE_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestCmpGTE_I8 = SSE_CmpGTE_I8; }
        else if (_avx2 ) { FastestCmpGTE_I8 = AVX2_CmpGTE_I8; }
        else { FastestCmpGTE_I8 = CPP_CmpGTE_I8; } // custom CPU implementation
    }
    FastestCmpGTE_I8(dst,src,bytes);
}
#pragma endregion

#pragma region Max
FunctionType3 FastestMax_U8 = NULL;
SSELIBRARY_API void Max_U8(char* dst, char* src, size_t bytes)
{
    if(FastestMax_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestMax_U8 = SSE_Max_U8; }
        //else if (_avx2 ) { FastestMax_U8 = AVX2_Max_U8; }
        else { FastestMax_U8 = CPP_Max_U8; }
    }
    FastestMax_U8(dst,src,bytes);
}

FunctionType3 FastestMax_I8 = NULL;
SSELIBRARY_API void Max_I8(char* dst, char* src, size_t bytes)
{
    if(FastestMax_I8 == NULL)
    {
        Init();
        if ( _sse41 ) { FastestMax_I8 = SSE41_Max_I8; }
        else if (_avx2 ) { FastestMax_I8 = AVX2_Max_I8; }
        else { FastestMax_I8 = CPP_Max_I8; }
    }
    FastestMax_I8(dst,src,bytes);
}
#pragma endregion

#pragma region Min
FunctionType3 FastestMin_U8 = NULL;
SSELIBRARY_API void Min_U8(char* dst, char* src, size_t bytes)
{
    if(FastestMin_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestMin_U8 = SSE_Min_U8; }
        else if (_avx2 ) { FastestMin_U8 = AVX2_Min_U8; }
        else { FastestMin_U8 = CPP_Min_U8; }
    }
    FastestMin_U8(dst,src,bytes);
}

FunctionType3 FastestMin_I8 = NULL;
SSELIBRARY_API void Min_I8(char* dst, char* src, size_t bytes)
{
    if(FastestMin_I8 == NULL)
    {
        Init();
        if ( _sse41 ) { FastestMin_I8 = SSE41_Min_I8; }
        else if (_avx2 ) { FastestMin_I8 = AVX2_Min_I8; }
        else { FastestMin_I8 = CPP_Min_I8; }
    }
    FastestMin_I8(dst,src,bytes);
}
#pragma endregion

#pragma region Diff
FunctionType3 FastestDiff_U8 = NULL;
SSELIBRARY_API void Diff_U8(char* dst, char* src, size_t bytes)
{
    if(FastestDiff_U8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestDiff_U8 = SSE_Diff_U8; }
        else if (_avx2 ) { FastestDiff_U8 = AVX2_Diff_U8; }
        else { FastestDiff_U8 = CPP_Diff_U8; }
    }
    FastestDiff_U8(dst,src,bytes);
}

FunctionType3 FastestDiff_I8 = NULL;
SSELIBRARY_API void Diff_I8(char* dst, char* src, size_t bytes)
{
    if(FastestDiff_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestDiff_I8 = SSE_Diff_I8; }
        else if (_avx2 ) { FastestDiff_I8 = AVX2_Diff_I8; }
        else { FastestDiff_I8 = CPP_Diff_I8; }
    }
    FastestDiff_I8(dst,src,bytes);
}
#pragma endregion

#pragma region Abs
FunctionType1 FastestAbs_I8 = NULL;
SSELIBRARY_API void Abs_I8(char* src, size_t bytes)
{
    if(FastestAbs_I8 == NULL)
    {
        Init();
        if ( _sse3 ) { FastestAbs_I8 = SSE_Abs_I8; }
        else if (_avx2 ) { FastestAbs_I8 = AVX2_Abs_I8; }
        else { FastestAbs_I8 = CPP_Abs_I8; }
    }
    FastestAbs_I8(src,bytes);
}
#pragma endregion

#pragma region Sign_I8
FunctionType3 FastestSign_I8 = NULL;
SSELIBRARY_API void Sign_I8(char* dst, char* src, size_t bytes)
{
    if(FastestSign_I8 == NULL)
    {
        Init();
        if ( _ssse3 ) { FastestSign_I8 = SSSE3_Sign_I8; }
        else if (_avx2 ) { FastestSign_I8 = AVX2_Sign_I8; }
        else { FastestSign_I8 = CPP_Sign_I8; }
    }
    FastestSign_I8(dst,src,bytes);
}
#pragma endregion

#pragma region BlendV_I8
FunctionType4 FastestBlendV_I8 = NULL;
SSELIBRARY_API void BlendV_I8(char* dst, char* src, char* mask, size_t bytes)
{
    if(FastestBlendV_I8 == NULL)
    {
        Init();
        if ( _ssse3 ) { FastestBlendV_I8 = SSE_BlendV_I8; }
        else if (_avx2 ) { FastestBlendV_I8 = AVX2_BlendV_I8; }
        else { FastestBlendV_I8 = CPP_BlendV_I8; }
    }
    FastestBlendV_I8(dst,src,mask,bytes);
}
#pragma endregion